/*
 * Device Driver for EG&G ACE-MCS 913
 *
 * Martin Gegenheimer, 1995,1996
 *
 *
 * $Id: eggmcs913.c,v 1.8 1996/06/30 14:21:45 martin Exp martin $
 * $Source: /home/martin/Software/DAQ/drivers/eggmcs913/RCS/eggmcs913.c,v $
 *
 *
 * $Log: eggmcs913.c,v $
 * Revision 1.8  1996/06/30 14:21:45  martin
 * Fake mode; master/multiple readers.
 *
 * Revision 1.7  1996/05/25 08:59:55  martin
 * Return true dwelltime, not the requested dwelltime.
 * Change macros to use base-address.
 * First try of multiple readers.
 *
 * Revision 1.6  1996/03/23  18:11:26  martin
 * read() now needs to be called with the # of bytes to transfer,
 * not with # of channels
 *
 * Revision 1.5  1996/03/21  10:17:37  martin
 * Changed all dprintk's  to MCS_PRINTK
 *
 * Revision 1.4  1996/03/13 09:26:19  martin
 * Fake driver operation
 *
 * Revision 1.3  1996/03/04  12:10:12  martin
 * Added error-message when not compiled as a module
 *
 * Revision 1.2  1996/02/22 15:42:27  martin
 * Changed eggmcs913_close() to eggmcs913_release
 *
 * Revision 1.1  1996/02/22  10:01:00  martin
 * Initial revision
 *
 *
 */

static char rcs_revision[] = "$Revision: 1.8 $";

#ifdef MODULE
#include <linux/module.h>
#else
#error "eggmcs913 has to be compiled as a module"
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/mman.h>

#include <linux/timer.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/string.h>


#if LINUX_VERSION_CODE >= 0x020000
#include <linux/wrapper.h>
#else
#error "check 'wrapper.h' for your linux-version"
#endif

#include "eggmcs913.h"

/* Only 1 board at the moment.

  static eggmcs913_t* boards[EGGMCS913_BOARDS_MAX]; */

static eggmcs913_t* boardPtr = NULL;

/* These routines allow a faked operation of the driver, to test
  applications, when the hardware in not present.

  A timer is used to generate a signal after a time corresponding to 
  the dwelltime. The channel-count is incremented, fake data is
  generated and if the passlength has been reached, an interrupt is
  simulated by calling the normal interrupt handler. */

static struct timer_list mcs_timer_list;
static int cur_channel = 0;

#if LINUX_VERSION_CODE <= 0x010300
static void eggmcs913_interrupt(int irq, struct pt_regs *ignored);
#elif LINUX_VERSION_CODE >= 0x020000
static void eggmcs913_interrupt(int irq, void* dontknow,
				struct pt_regs *ignored);
#else
#error "check 'irqhandler-type for your linux-version"
#endif
static void mcs_start_timer(void);

static void eggmcs913_timer_interrupt(unsigned long unused)
{
	unsigned long *b = (unsigned long*) boardPtr->membase;
	int i;

	del_timer(&mcs_timer_list);

	if (cur_channel == boardPtr->passlength) {
		cur_channel = 0;
		if (boardPtr->add==0) {
			for (i=0; i<boardPtr->nchannels; i++) {
				b[i] = 0;
			}
		}
#if LINUX_VERSION_CODE <= 0x010300
		eggmcs913_interrupt(boardPtr->irq, NULL);
#elif LINUX_VERSION_CODE >= 0x020000
		eggmcs913_interrupt(boardPtr->irq, NULL, NULL);
#else
#error "check 'irqhandler-type for your linux-version"
#endif
	} else {
		b[cur_channel]+=cur_channel%50+1;
		cur_channel++;
		mcs_start_timer();
	}
}

static void mcs_start_timer(void)
{

	if (boardPtr->active == 0) {
		cur_channel = 0;
		return;
	}
	init_timer(&mcs_timer_list);
	mcs_timer_list.expires = jiffies + 2*(boardPtr->dwelltime+1);
	mcs_timer_list.function = eggmcs913_timer_interrupt;
	add_timer(&mcs_timer_list);
}

/* end of fake routines */

static int calc_timers(unsigned int dwt, int* timer1, int* timer2,
		       int* remain)
{
	unsigned int  tmax = 0x10000;
	int tmin = 2;
	int f, r;

	int rmin;

	int i, imin;

	unsigned int ticks;


	ticks = 2*dwt;

	if (dwt == 0) {
		*timer1 = 0;
		*timer2 = 0;
		*remain = 0;
		return 0;
	}
	if (dwt < EGGMCS913_MIN_DWELLTIME) {
		printk("dwelltime = %d < %d, not allowed \n",\
		       dwt, boardPtr->mindwelltime);
		return -EINVAL;
	}
	if (dwt > 2147418112) {
		printk("dwelltime = %d > 2147418112, not allowed \n", dwt);
		return -EINVAL;
	}
	if (dwt <tmax) {
		*timer1 = dwt;
		*timer2 = 2;
		*remain = 0;
		return 0;
	} else {
		rmin = tmax;
		imin = tmin;
		for (i = ticks/(tmax-1)+1; i< tmax; i++) {
			r = ticks%i;
			if (r<=rmin) {
				f = ticks/i;
				rmin = r;
				imin = i;
			}
		}

		*timer2 = ticks/imin;
		*timer1 = imin;
		*remain = ticks%imin;
        
		return 0;
	}
}

static void eggmcs913_board_select(int board)
{
	MCS_PRINTK("eggmcs913_board_select: %d.\n", board);

	/* select board */
	outb_p(board, EGGMCS913_ADDRESS(boardPtr->baseaddr) );
	/* init, and reset it  ??? */
	outb_p(EGGMCS913_UNKNOWN_CMD, EGGMCS913_CONTROL(boardPtr->baseaddr) );
}

/* Don't know when exactly that changed */
#if LINUX_VERSION_CODE <= 0x010300
EGGMCS913_EXTERN int eggmcs913_set_irqhandler(void irqhndl(int irq,
 struct pt_regs *ignored)) 
{
	if (boardPtr->irqhandler != NULL) {
		free_irq(boardPtr->irq);
		boardPtr->irqhandler = NULL;
	}

	if(request_irq(boardPtr->irq, irqhndl, SA_INTERRUPT, "eggmcs913")) {
		printk("eggmcs913: set_irqhandler: Cannot request IRQ(%d).\n",
		       boardPtr->irq);
		return -(EIO);
	}

	boardPtr->irqhandler = irqhndl;

	return 0;

}
#elif LINUX_VERSION_CODE >= 0x020000
EGGMCS913_EXTERN int eggmcs913_set_irqhandler(void irqhndl(int irq, void*,
 struct pt_regs *ignored)) 
{
	if (boardPtr->irqhandler != NULL) {
		free_irq(boardPtr->irq, NULL);
		boardPtr->irqhandler = NULL;
	}

	if(request_irq(boardPtr->irq, irqhndl, SA_INTERRUPT, "eggmcs913",
		       NULL)) {
		printk("eggmcs913: set_irqhandler: Cannot request IRQ(%d).\n",
		       boardPtr->irq);
		return -(EIO);
	}

	boardPtr->irqhandler = irqhndl;

	return 0;
}

#else
#error "check 'free_irq/request_irq'-call for your linux-version"
#endif

#ifdef EGGMCS913_EXP 
EGGMCS913_EXTERN void eggmcs913_set_params(int board, 
					   long dwelltime,
					   int passlength,
					   unsigned long passcountpreset,
					   int extstart,
					   int extdwelltime,
					   int add
					   )
{
	MCS_PRINTK("eggmcs913_set_params: %d.\n", board);

	boardPtr->dwelltime       = dwelltime;
	boardPtr->passlength      = passlength;
	boardPtr->passcountpreset = passcountpreset;
	boardPtr->extstart        = extstart;
	boardPtr->extdwelltime    = extdwelltime;
	boardPtr->add             = add;

	boardPtr->passcount       = 0;
	boardPtr->active          = 0;
	boardPtr->cmd             = 0;
	boardPtr->stop_req        = 0;

	MCS_PRINTK("eggmcs913_set_params: dt %ld.\n", boardPtr->dwelltime);
}

#endif /* EGGMCS913_EXP */


/* Set the default values. Only in the struct, the board is
   programmed  when started. */
static void eggmcs913_set_default(int board)
{
	MCS_PRINTK("eggmcs913_default: %d.\n", board);

	boardPtr->dwelltime       = 250;
	boardPtr->passlength      = 4;
	boardPtr->passcountpreset = 1;
	boardPtr->extstart        = 0;
	boardPtr->extdwelltime    = 0;
	boardPtr->add             = 1;

	boardPtr->passcount       = 0;
	boardPtr->cmd             = 0;
}

/*
  set_dwelltime
 
  dwelltime in usec.
 
  Program the counters for the dwelltime; there are 2 cascaded
  2-byte * counters. Their mode is */
static int eggmcs913_set_dwelltime(long dwelltime)
{
	int cnt0, cnt1;

	MCS_PRINTK("eggmcs913_set_dwelltime: %d us = 0x%04x.\n",
		   dwelltime, dwelltime); 

	cnt0 = boardPtr->cnt0;
	cnt1 = boardPtr->cnt1;


	MCS_PRINTK("eggmcs913_set_dwelltime: cnt0 = %d  cnt1 = %d\n",
		   cnt0, cnt1);

	/* for 0xffff the counter has to be set to 0 */
	if (cnt0 == 0xffff) cnt0 = 0;
	if (cnt1 == 0xffff) cnt1 = 0;

	/* this should never happen */
	if (cnt0 < 2  || cnt1 < 2) {
		printk( "eggmcs913_set_dwelltime: ERROR cnt <2 !!!!!\n" );
		return -1;
	}

	/* set counters */

	/* counter 1 (the write operations to control-word addressing
	   counter0 with different modes make the out0 go high/low
	   which is connected to clk1. This is necessary to
	   make cnt1 load the presets). */

	/* 0x30 = 001100110 :  counter0, twobyte transfer, mode 0 */
	outb_p( 0x30, EGGMCS913_TIM8253_CW(boardPtr->baseaddr) );
	/* 0x74 = 011100110 :  counter1, twobyte transfer, mode 2 */
	outb_p( 0x74, EGGMCS913_TIM8253_CW(boardPtr->baseaddr) );

	MCS_PRINTK("eggmcs913_set_dwelltime: cnt1 low  = 0x%02x\n",
		   LOWBYTE(cnt1));
	MCS_PRINTK("eggmcs913_set_dwelltime: cnt1 high = 0x%02x\n",
		   HIGHBYTE(cnt1));
	outb_p(  LOWBYTE(cnt1), EGGMCS913_TIM8253_C1(boardPtr->baseaddr) );
	outb_p( HIGHBYTE(cnt1), EGGMCS913_TIM8253_C1(boardPtr->baseaddr) );

	/* 0x34 = 001100100 :  counter0, twobyte transfer, mode 2 */
	outb_p( 0x34, EGGMCS913_TIM8253_CW(boardPtr->baseaddr) );
	/* 0x30 = 001100000 :  counter0, twobyte transfer, mode 0 */
	outb_p( 0x30, EGGMCS913_TIM8253_CW(boardPtr->baseaddr) );


	/* counter 0 */
	/* 0x34 = 001100100 :  counter0, twobyte transfer, mode 2 */
	outb_p( 0x34, EGGMCS913_TIM8253_CW(boardPtr->baseaddr) );

	MCS_PRINTK("eggmcs913_set_dwelltime: cnt0 low  = 0x%02x\n",
		   LOWBYTE(cnt0));
	MCS_PRINTK("eggmcs913_set_dwelltime: cnt0 high = 0x%02x\n",
		   HIGHBYTE(cnt0));
	outb_p( LOWBYTE(cnt0), EGGMCS913_TIM8253_C0(boardPtr->baseaddr));
	outb_p(HIGHBYTE(cnt0), EGGMCS913_TIM8253_C0(boardPtr->baseaddr));

	return 0;
}

static void eggmcs913_set_passlength(int passlength)
{
	MCS_PRINTK("eggmcs913_set_passlength: %d.\n", passlength);
	/* 0xb6 = 101100110 :  counter2, twobyte transfer, mode 3 */
	outb_p(0xb6, EGGMCS913_TIM8253_CW(boardPtr->baseaddr));

	outb_p( LOWBYTE(passlength), EGGMCS913_TIM8253_C2(boardPtr->baseaddr));
	outb_p(HIGHBYTE(passlength), EGGMCS913_TIM8253_C2(boardPtr->baseaddr));

}

static void eggmcs913_clear(void)
{
	int i;
	unsigned long *b = (unsigned long*) boardPtr->membase;

	MCS_PRINTK("eggmcs913_clear:.\n");

	for (i=0; i<boardPtr->nchannels; i++) 
		{
			b[i] = 0;
			/* b[i] = i;  */
		}

	boardPtr->passcount       = 0;

}

static void eggmcs913_makecmd(void)
{
	int cmd;

	MCS_PRINTK("eggmcs913_makecmd:.\n");

	cmd = 0;

	if (boardPtr->add) cmd += EGGMCS913_ADDMODE;
	if (boardPtr->extdwelltime)
		{
			cmd += EGGMCS913_EXTDWELLTIME;
		}

	if (boardPtr->extstart) cmd += EGGMCS913_EXTSTART;

	MCS_PRINTK("eggmcs913_makecmd:cmd = 0x%02x.\n", cmd);
	boardPtr->cmd = cmd;

}

EGGMCS913_EXTERN void eggmcs913_restart(void)
{
	MCS_PRINTK("eggmcs913_restart:.\n");

	if (boardPtr->extdwelltime) {
		/* 0x70 = 011100000 :  counter1, twobyte transfer,
		   mode 0 => OUT1=LOW */
		outb_p( 0x70, EGGMCS913_TIM8253_CW(boardPtr->baseaddr) );
	}

	MCS_PRINTK("eggmcs913_restart:cmd = 0x%02x.\n", boardPtr->cmd);

	outb_p( boardPtr->cmd|EGGMCS913_UNKNOWN_BIT,
		EGGMCS913_CONTROL(boardPtr->baseaddr) );
    
	if (boardPtr->fake != 0) {
		if (boardPtr->active == 1) {
			mcs_start_timer();
		}
		return;
	}
}

EGGMCS913_EXTERN int eggmcs913_start(void)
{
	int cmd, stat = 0;

	MCS_PRINTK("eggmcs913_start:.\n");

	if (boardPtr->active == 1) {
		printk("eggmcs913_start: already running.\n");
		return 0;
	}

	if (boardPtr->passcount == boardPtr->passcountpreset) {
		MCS_PRINTK("eggmcs913: passcount == passcountpreset\n");
		return 0;
	}

	cmd = 0;

	if (boardPtr->add) cmd += EGGMCS913_ADDMODE;
	if (boardPtr->extdwelltime) {
		cmd += EGGMCS913_EXTDWELLTIME;
		/* 0x70 = 011100000 :  counter1, twobyte transfer,
		   mode 0 => OUT1=LOW */
		outb_p( 0x70, EGGMCS913_TIM8253_CW(boardPtr->baseaddr) );
	} else {
		if ((stat = eggmcs913_set_dwelltime(boardPtr->dwelltime)) <0) {
			return stat;
		}
	}

	eggmcs913_set_passlength(boardPtr->passlength);

	if (boardPtr->extstart) cmd += EGGMCS913_EXTSTART;

	MCS_PRINTK("eggmcs913_start:cmd = 0x%02x.\n", cmd);

	eggmcs913_makecmd();

	outb_p( cmd                      ,
		EGGMCS913_CONTROL(boardPtr->baseaddr) );
	outb_p( cmd|EGGMCS913_UNKNOWN_BIT,
		EGGMCS913_CONTROL(boardPtr->baseaddr) );
	outb_p( EGGMCS913_UNKNOWN_CMD    ,
		EGGMCS913_CONTROL(boardPtr->baseaddr) );
	outb_p( cmd|EGGMCS913_UNKNOWN_BIT,
		EGGMCS913_CONTROL(boardPtr->baseaddr) );


	boardPtr->active = 1;

	if (boardPtr->fake != 0) {
		mcs_start_timer();
	}

	return 0;
}

static int eggmcs913_status(void)
{
	int status;

	status = inb_p(EGGMCS913_CONTROL(boardPtr->baseaddr) );
	MCS_PRINTK("eggmcs913_status: 0x%02x.\n", status);
	return status;
}

EGGMCS913_EXTERN void eggmcs913_stop(void)
{
	MCS_PRINTK("eggmcs913_stop:.\n");
	outb_p( EGGMCS913_UNKNOWN_CMD, EGGMCS913_CONTROL(boardPtr->baseaddr) );

	/* It should be checked wether the board is really
	   deactivated */
	boardPtr->active = 0;
}

#if LINUX_VERSION_CODE <= 0x010300
static void eggmcs913_interrupt(int irq, struct pt_regs *ignored)
#elif LINUX_VERSION_CODE >= 0x020000
	static void eggmcs913_interrupt(int irq, void* dontknow,
					struct pt_regs *ignored)
#else
#error "check 'irqhandler-type for your linux-version"
#endif
{
	int status;

	MCS_PRINTK("eggmcs913: interrupted.\n");

	boardPtr->passcount++;

	status = inb_p(EGGMCS913_CONTROL(boardPtr->baseaddr) );
	if (boardPtr->fake != 0) {
		status = 0;
	}

	MCS_PRINTK("eggmcs913: passcount = %ld.\n", boardPtr->passcount);
	MCS_PRINTK("eggmcs913: passcountpreset = %ld.\n",
		   boardPtr->passcountpreset);

	if (status & EGGMCS913_EXTSTOP) {
		MCS_PRINTK("eggmcs913: interrupt, ext stop requested\n");
		eggmcs913_stop();
		return;
	}
    
	if (boardPtr->stop_req) {
		MCS_PRINTK("eggmcs913: interrupt, stop requested\n");
		eggmcs913_stop();
		return;
	}
	if (boardPtr->passcount == boardPtr->passcountpreset) {
		MCS_PRINTK("eggmcs913: passcount == passcountpreset\n");
		eggmcs913_stop();
	} else {
		outb_p(EGGMCS913_UNKNOWN_CMD,
		       EGGMCS913_CONTROL(boardPtr->baseaddr) );
		eggmcs913_restart();
	}
}

static int eggmcs913_read(struct inode *inode, struct file *filp,
                          char *buffer, int count)
{
	EGGMCS913_ClientData* mcs = (EGGMCS913_ClientData*) filp->private_data;
	int stat;
	/*
	  const unsigned long mask = 0xFFFFFFL;
	  */
	char* b = (char*) boardPtr->membase;
	char* bb;

	MCS_PRINTK("eggmcs913_read: reading out 0x%x bytes from 0x%lx.\n",
		   count, boardPtr->membase);


	bb = (char *) b;
	bb += (EGGMCS913_BYTESPERCHANNEL) * mcs->startchannel;
      
	if (count<0) return -(EINVAL);
         
	if (bb+count>b+boardPtr->memsize) return -(EINVAL);
	/*
	  if (bb+count>b+boardPtr->memsize) count = boardPtr->memsize;
	  */
         
	if ( (stat = verify_area(VERIFY_WRITE, (char *)buffer, count )) != 0) {
		return stat;
	}
	memcpy_tofs(buffer, bb, count);
    
	return count;
}

static unsigned long eggmcs913_malloc(struct file *filp, unsigned long size);
static int eggmcs913_free(struct file *filp, unsigned long ptr);

static int eggmcs913_ioctl(struct inode *inode, struct file *filp,
                           unsigned int cmd, unsigned long arg)
{
	unsigned int stat;
	EGGMCS913_ClientData* mcs = (EGGMCS913_ClientData*) filp->private_data;

	switch (cmd)
		{
		case EGGMCS913_IOCTL_SET_DWELLTIME: {
			int cnt0, cnt1, remain;
			int dwelltime = (int) arg;

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			if (dwelltime < EGGMCS913_MIN_DWELLTIME ||
			    dwelltime > EGGMCS913_MAX_DWELLTIME ) {
				return -(EINVAL);
			}
			boardPtr->dwelltime = dwelltime;
			if ( (stat = calc_timers(dwelltime, &cnt0, &cnt1,
						 &remain)) < 0) {
				printk( "eggmcs913_set_dwelltime: "
					"ERROR cnt <2 !!!!!\n" );
				return stat;
			}
			boardPtr->cnt0 = cnt0;
			boardPtr->cnt1 = cnt1;
		}
		break;


		case EGGMCS913_IOCTL_SET_PASSLENGTH: {
			int passlength = (int) arg;

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			if (passlength < EGGMCS913_MIN_PASSLENGTH ||
			    passlength > EGGMCS913_MAX_PASSLENGTH ) {
				return -(EINVAL);
			}
			boardPtr->passlength = passlength;
		}
		break;

		case EGGMCS913_IOCTL_SET_PASSCOUNTPRESET: {
			unsigned long passcountpreset = (unsigned long) arg;

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			if ( passcountpreset < 1 ) {
				return -(EINVAL);
			}
			boardPtr->passcountpreset = passcountpreset;
		}
		break;


		case EGGMCS913_IOCTL_START:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			stat = eggmcs913_start();
			return stat;
			break;
		case EGGMCS913_IOCTL_STOP:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			eggmcs913_stop();
			break;
		case EGGMCS913_IOCTL_STOPREQ:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->stop_req = (int)arg;
			break;

		case EGGMCS913_IOCTL_CLEAR:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			eggmcs913_clear();
			break;

		case EGGMCS913_IOCTL_SET_EXTDWELLTIME:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->extdwelltime = (int)arg;
			break;
		case EGGMCS913_IOCTL_SET_EXTSTART:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->extstart = arg;
			break;
		case EGGMCS913_IOCTL_SET_ADD:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->add = arg;
			break;



		case EGGMCS913_IOCTL_GET_DWELLTIME: {
			int dwelltime = boardPtr->cnt0 * boardPtr->cnt1 /2;
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(dwelltime, (long *) arg);
		}
		break;
		case EGGMCS913_IOCTL_ACTIVE:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardPtr->active, (long *) arg);
			break;
		case EGGMCS913_IOCTL_GET_PASSCOUNT:
			if ((stat = verify_area(VERIFY_WRITE, (char *)arg,
						sizeof(unsigned long))) != 0){
				return stat;
			}

			put_user(boardPtr->passcount, (unsigned long *)arg);
			break;
		case EGGMCS913_IOCTL_GET_EXTDWELLTIME:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0) {
				return stat;
			}

			put_user(boardPtr->extdwelltime, (long *) arg);
			break;
		case EGGMCS913_IOCTL_GET_EXTSTART:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardPtr->extstart, (long *) arg);
			break;
		case EGGMCS913_IOCTL_GET_ADD:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardPtr->add, (long *) arg);
			break;

		case EGGMCS913_IOCTL_NCHANNELS:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardPtr->nchannels, (long *) arg);
			break;

		case EGGMCS913_IOCTL_STARTCHANNEL:
			mcs->startchannel = (int)  arg;
			break;

		case EGGMCS913_IOCTL_STATUS:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(eggmcs913_status(), (long *) arg);
			break;

		case EGGMCS913_IOCTL_MMAP:
			return eggmcs913_malloc(filp, (unsigned long) arg);
			break;

		case EGGMCS913_IOCTL_MUNMAP:
			return eggmcs913_free(filp, (unsigned long) arg);
			break;

		default: 
			MCS_PRINTK("eggmcs913: unknown IOCTL: %d\n.", cmd);
			return -(EINVAL);
			break;
		}

	return 0;
}

static int eggmcs913_mmap(struct inode *inode, struct file *filp,
			  struct vm_area_struct *vma)
{
	printk("eggmcs913: offset = %ld\n", vma->vm_offset);

	if (remap_page_range(vma->vm_start, vma->vm_offset,
			     vma->vm_end - vma->vm_start,
			     vma->vm_page_prot)) {
		printk("eggmcs913: remap bad\n");
		return -EAGAIN;
	}

	printk("eggmcs913: remap ok\n");
	/* vma->vm_inode = inode;  XXX */
	/* inode->i_count++; */
	return 0;
}

static unsigned long eggmcs913_malloc(struct file *filp, unsigned long size)
{
	unsigned long uAdr, pAdr;

	struct file_operations fops;
	struct file memfile;

	printk("eggmcs913: size = %ld\n", size);

	fops.mmap = eggmcs913_mmap;
	memfile.f_op = &fops;
	memfile.f_mode = 0x3;

	/* check size */
	if (size <= PAGE_SIZE/2) {
		size = PAGE_SIZE - 0x10;
	}
	pAdr = (unsigned long) boardPtr->membase;

	for (uAdr = pAdr & PAGE_MASK; uAdr < pAdr+size; uAdr += PAGE_SIZE) {
#if LINUX_VERSION_CODE < 0x01031D
		mem_map[MAP_NR(uAdr)] |= MAP_PAGE_RESERVED; /* < 1.3.29 */
#elif LINUX_VERSION_CODE < 0x01033A                 /* < 1.3.58 */
		mem_map[MAP_NR(uAdr)].reserved = 1;
#else
		mem_map_reserve(MAP_NR(uAdr));
#endif
	}

	uAdr = do_mmap(&memfile, 0, (size + ~PAGE_MASK) & PAGE_MASK,
		       PROT_READ|PROT_WRITE|PROT_EXEC, MAP_SHARED,
		       pAdr&PAGE_MASK);

	if ((signed long) uAdr <= 0) {
		printk("do_mmap failed\n");
		return uAdr;
	}

	uAdr |= pAdr & ~PAGE_MASK;
	return uAdr;
}
static int eggmcs913_free(struct file *filp, unsigned long ptr)
{
	unsigned long size = 0;

	/* XXX */
	do_munmap(ptr & PAGE_MASK, (size+(~PAGE_MASK)) & PAGE_MASK);

	do {
#if LINUX_VERSION_CODE < 0x01031D
		mem_map[MAP_NR(ptr)] &= ~MAP_PAGE_RESERVED;
#elif LINUX_VERSION_CODE < 0x01033A
		mem_map[MAP_NR(ptr)].reserved = 0;
#else
		mem_map_unreserve(MAP_NR(ptr));
#endif

		ptr += PAGE_SIZE;
	} while (0);

	return 0;
}

/*
   This may be called by another driver
*/
EGGMCS913_EXTERN int eggmcs913_open(struct inode *inode, struct file *filp)
{
	int mode;
	int controler;
	EGGMCS913_ClientData* mcs;

	mode = filp->f_flags & O_ACCMODE;

	if (boardPtr->controler == 1) {
		if (mode==O_RDWR) {
			return -(EBUSY);
		}
		if (mode!=O_RDONLY) {
			return -(EINVAL);
		}
		controler = 0;
	} else {
		if (mode!=O_RDONLY) {
			controler = 1;
		} else {
			controler = 0;
		}
	}

	mcs = (EGGMCS913_ClientData*) kmalloc(sizeof(EGGMCS913_ClientData),
					      GFP_KERNEL);
	if (mcs == NULL) {
		return -(ENOMEM);
	}
	filp->private_data = mcs;

	mcs->controler    = controler;
	mcs->startchannel = 0;

	if (mcs->controler == 1) {
		if (eggmcs913_set_irqhandler(eggmcs913_interrupt) != 0) {
			printk("eggmcs913: open: "
			       "Could not install irq-handler.\n");
			return -(EINVAL);
		}
		eggmcs913_set_default(0);
		eggmcs913_board_select(0);
		boardPtr->controler = controler;
	}

	MOD_INC_USE_COUNT;

	boardPtr->usecount++;
	MCS_PRINTK("eggmcs913_open: EGGMCS913 opened.\n");

	return 0;
}

void eggmcs913_release(struct inode *inode, struct file *filp)
{
	EGGMCS913_ClientData* mcs = (EGGMCS913_ClientData*) filp->private_data;

	if (boardPtr->active==1 && boardPtr->usecount == 1) {
		eggmcs913_stop();
		if (boardPtr->fake != 0) {
			/* to get rid of a running timer */
			eggmcs913_timer_interrupt(0); 
			cur_channel = 0;
		}
	}

	if (mcs->controler==1) {
		boardPtr->controler = 0;
	}

	kfree(mcs);

	MOD_DEC_USE_COUNT;
	boardPtr->usecount--;

	MCS_PRINTK("eggmcs913_release: usecount = %d.\n", boardPtr->usecount);
}

struct file_operations eggmcs913_fops = {
	NULL,               /* seek */
	eggmcs913_read,     /* read */
	NULL,               /* write */
	NULL,               /* readdir */
	NULL,               /* select */
	eggmcs913_ioctl,    /* ioctl */
	NULL,               /* mmap */
	eggmcs913_open,     /* open */
	eggmcs913_release   /* release */
};


#ifdef MODULE    /* only module version */

#if LINUX_VERSION_CODE <= 0x010300
char kernel_version[] = UTS_RELEASE;
#elif LINUX_VERSION_CODE >= 0x020000
#else
#error "check 'kernel_version' - symbol for your linux-version"
#endif

/* These can be set from the commandline, when loading the module. */
static int major = EGGMCS913_MAJOR;

static int irq = EGGMCS913_IRQ;
static int io = EGGMCS913_BASEADDR;
static int mem = EGGMCS913_MEMBASE;

static int fake = 0;


/* Name and version # of the driver, the name is used be load-script
   to find a dynamically assigned major device # (one word) */
static char* eggmcs913_name = "eggmcs913-0.4.0";


int init_module(void)
{
	int err;

	printk("eggmcs913: EG&G ACE-MCS 913 driver, version %s\n",
	       rcs_revision);

	boardPtr = (eggmcs913_t *)kmalloc( sizeof(eggmcs913_t), GFP_KERNEL);
	if (boardPtr == NULL) {
		return -(ENOMEM);
	}

	boardPtr->fake            = 0;

	boardPtr->usecount        = 0;
	boardPtr->controler       = 0;

	boardPtr->stop_req        = 0;
	boardPtr->active          = 0;

	boardPtr->irq             = irq;
	boardPtr->baseaddr        = io;
	boardPtr->membase         = (unsigned long*) mem;

	boardPtr->fake            = fake;

	boardPtr->bytesperchannel = EGGMCS913_BYTESPERCHANNEL;
	boardPtr->nchannels       = EGGMCS913_CHANNELS;
	boardPtr->memsize         =
		boardPtr->bytesperchannel*boardPtr->nchannels;

	boardPtr->irqhandler      = NULL;

	if (boardPtr->fake != 0) {
		boardPtr->membase = (unsigned long *)
		      kmalloc( boardPtr->memsize, GFP_KERNEL);
		if (boardPtr->membase == NULL) {
			return -(ENOMEM);
		}
		eggmcs913_clear();
	}

	if (check_region(boardPtr->baseaddr, EGGMCS913_IOSIZE) != 0) {
		printk("eggmcs913: init_module: "
		       "I/O ports 0x%lx to 0x%lx already in use.\n",
		       boardPtr->baseaddr,
		       boardPtr->baseaddr + EGGMCS913_IOSIZE-1);
		goto fail_region;
	}

	request_region(boardPtr->baseaddr, EGGMCS913_IOSIZE, "eggmcs913");

	/* How to reserve the memory at  boardPtr->membase ? */


	err = register_chrdev(major, eggmcs913_name, &eggmcs913_fops);
	if (err < 0) {
		printk("eggmcs913: init_module: "
		       "Cannot register to major device %d.\n", major);
		goto fail_major;
	}
	if (major == 0) {
		major = err; /* dynamic major # */
	}

	printk("eggmcs913: installed.\n");
	printk("eggmcs913: init_module: "
	       "fake(%d), major(%d), irq(%d), baseaddr(0x%lx), "
	       "membase(0x%p).\n",
	       boardPtr->fake, major, boardPtr->irq, boardPtr->baseaddr,
	       boardPtr->membase);

	return 0;

	printk("eggmcs913: unloading driver.\n");
fail_major:
	release_region(boardPtr->baseaddr, EGGMCS913_IOSIZE);
fail_region:
	kfree(boardPtr);

	return -(EIO);
}

void cleanup_module(void)
{
	if (MOD_IN_USE) {
		printk("eggmcs913: device busy, remove delayed\n");
		return;
	} else {
		if (boardPtr->irqhandler != NULL) {
#if LINUX_VERSION_CODE <= 0x010300
			free_irq(boardPtr->irq);
#elif LINUX_VERSION_CODE >= 0x020000
			free_irq(boardPtr->irq, NULL);
#else
#error "check 'free_irq'-call for your linux-version"
#endif
		}
		release_region(boardPtr->baseaddr, EGGMCS913_IOSIZE);

		if (unregister_chrdev(major, eggmcs913_name) != 0) {
			printk("eggmcs913: cleanup_module failed\n");
		} else {
			MCS_PRINTK("eggmcs913: cleanup_module succeeded\n");
		}
		if (boardPtr->fake != 0) {
			kfree(boardPtr->membase);
		}
		kfree(boardPtr);
	}
}
#endif /* MODULE */
